#include <gmc/gmc.h>
#include <stdio.h>
#include <slibs/slibs.h>
#include <zlib.h>

#define GMC_WM(file, variable, count) fwrite(&variable, sizeof(variable), count, file)
#define GMC_RM(file, variable, count) fread(&variable, sizeof(variable), count, file)
#define GMC_W(file, variable) GMC_WM(file, variable, 1)
#define GMC_R(file, variable) GMC_RM(file, variable, 1)

gmc_file gmc_read(const char* filename) {
    FILE* input = fopen(filename, "rb");

    gmc_file file = { 0 };
    GMC_R(input, file.magic);
    GMC_R(input, file.string_buf_len);
    file.string_buf = calloc(file.string_buf_len, sizeof(char));
    GMC_RM(input, *file.string_buf, file.string_buf_len);
    GMC_R(input, file.entity_count);
    file.entities = calloc(file.entity_count, sizeof(gmc_entity));
    for(uint32_t i = 0; i < file.entity_count; i++) {
        gmc_entity* entity = &file.entities[i];
        GMC_R(input, entity->property_count);
        entity->properties = calloc(entity->property_count, sizeof(gmc_property));
        GMC_RM(input, *entity->properties, entity->property_count);
    }

    return file;
}

gmc_file gmc_from_map(map_map map) {
    sl_string strings = { 0 };
    sl_map(string_indicies, int);

    sl_vec(gmc_entity) gmc_ents = { 0 };
    for(sl_vec_it(entity, map.entities)) {
        sl_vec(gmc_property) gmc_props = { 0 };

        for(sl_vec_it(property, entity->properties)) {
            gmc_property gmc_prop = { 0 };
            if(sl_map_get(string_indicies, property->key) == NULL) {
                gmc_prop.key = strings.size;

                sl_append_c_str(&strings, property->key);
                sl_vec_push(strings, '\0');

                sl_map_set(string_indicies, property->key, gmc_prop.key);
            } else {
                gmc_prop.key = *sl_map_get(string_indicies, property->key);
            }

            if(sl_map_get(string_indicies, property->value) == NULL) {
                gmc_prop.value = strings.size;

                sl_append_c_str(&strings, property->value);
                sl_vec_push(strings, '\0');

                sl_map_set(string_indicies, property->value, gmc_prop.value);
            } else {
                gmc_prop.value = *sl_map_get(string_indicies, property->value);
            }

            sl_vec_push(gmc_props, gmc_prop);
        }

        gmc_entity gmc_ent = {
            .property_count = gmc_props.size,
            .properties = gmc_props.data
        };
        sl_vec_push(gmc_ents, gmc_ent);
    }

    gmc_file file = {
        .magic = gmc_magic,

        .string_buf_len = strings.size,
        .string_buf = strings.data,

        .entity_count = gmc_ents.size,
        .entities = gmc_ents.data
    };

    return file;
}

void gmc_write(gmc_file file, const char* filename) {
    FILE* output = fopen(filename, "wb");

    GMC_W(output, file.magic);
    
    GMC_W(output, file.string_buf_len);
    fwrite(file.string_buf, sizeof(char), file.string_buf_len, output);

    GMC_W(output, file.entity_count);
    for(uint32_t i = 0; i < file.entity_count; i++) {
        gmc_entity entity = file.entities[i];

        GMC_W(output, entity.property_count);
        for(uint32_t j = 0; j < entity.property_count; j++) {
            gmc_property property = entity.properties[j];

            GMC_W(output, property.key);
            GMC_W(output, property.value);
        }
    }

    fclose(output);
}

void gmc_print(gmc_file file) {
    printf("--GMC File--\n");
    printf("Magic: %llu\n\n", file.magic);
    
    printf("String buffer length: %d\n", file.string_buf_len);
    printf("String buffer:\n0: ");
    for(uint32_t i = 0; i < file.string_buf_len; i++) {
        char ch = file.string_buf[i];
        if(ch == '\0') {
            printf("\n%d: ", i + 1);
        } else {
            printf("%c", ch);
        }
    }

    printf("\n\nEntity count: %d\n", file.entity_count);
    for(uint32_t i = 0; i < file.entity_count; i++) {
        gmc_entity entity = file.entities[i];
        printf("Entity %d:\n", i);
        printf("\tProperty count: %d\n", entity.property_count);
        printf("\tProperties:\n");
        for(uint32_t j = 0; j < entity.property_count; j++) {
            gmc_property property = entity.properties[j];
            printf("\t\t%d (%s): %d (%s)\n", property.key, &file.string_buf[property.key], 
                   property.value, &file.string_buf[property.value]);
        }
    }
}

const uint64_t gmc_magic = 
    ((uint64_t)('G') << 0 | (uint64_t)('M') << 8 | (uint64_t)('C') << 16 |
     (uint64_t)('F') << 24 | (uint64_t)('M') << 32 | (uint64_t)('T') << 40 |
     (uint64_t)('0') << 48 | (uint64_t)('1') << 56);
